import pipes
import subprocess

from typing import List, Optional, Tuple, Union

from edge.types.exceptions import ExternalError, InternalError, ToutError


def run_silent(esc_cmd: str) -> int:
    """
    :param esc_cmd:
        Escaped shell-command. Use pipes.quote() to escape input form external sources
    """
    return subprocess.run(esc_cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT).returncode


def quote(data: str) -> str:
    return pipes.quote(data)


def exec_cli(cmd: str, args: Union[List[str], Tuple[str]],
             tout_sec: Optional[int] = 10,
             encoding: str = 'ascii') -> str:
    """
    :param cmd:
    :param args:
    :param tout_sec:
    :param encoding: 'ascii' | 'utf-8'
    :returns: stdout+stderr
    """
    from edge import eventlog
    eventlog.log_d(f'exec: <{cmd}>')

    from pexpect.popen_spawn import PopenSpawn
    from pexpect.exceptions import EOF, TIMEOUT

    child_cmd = ' '.join([cmd] + args)
    if encoding != 'ascii':
        child_cmd = child_cmd.encode(encoding)

    child = PopenSpawn(child_cmd, timeout=tout_sec, encoding=encoding)

    try:
        child.expect_exact(EOF)
        child.wait()
    except TIMEOUT:
        eventlog.log_e(f'fail: ToutError (timeout:{tout_sec}sec)')
        raise ToutError

    if child.exitstatus is None:
        eventlog.log_e(f'fail: InternalError (signal-status:{child.signalstatus})')
        raise InternalError

    elif child.exitstatus == 0:
        eventlog.log_d('done')
        return child.before

    elif child.exitstatus == 64:
        eventlog.log_e('fail: ExternalError (exit-status:EX_USAGE)')
        raise ExternalError

    elif child.exitstatus == 70:
        eventlog.log_e(f'fail: InternalError (exit-status:EX_SOFTWARE, before:<{child.before}>)')
        raise InternalError
