import fcntl
import os

from typing import List, Set, Union

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import ph


def try_lock(lock_dir: str, lock_name: str) -> int:
    import warnings
    warnings.warn('message', PendingDeprecationWarning, stacklevel=2)

    if not os.path.exists(lock_dir):
        os.makedirs(lock_dir)

    lock_path = fs.join_path(lock_dir, lock_name + '.lock')
    lock_fd = os.open(lock_path, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
    try:
        # Request exclusive (EX) non-blocking (NB) advisory lock.
        fcntl.lockf(lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        return lock_fd

    except IOError as x:
        eventlog.log_e(f'fail: fcntl.lockf ({x})')
        raise BusyError


def unlock(lock_fd):
    import warnings
    warnings.warn('message', PendingDeprecationWarning, stacklevel=2)
    try:
        fcntl.lockf(lock_fd, fcntl.LOCK_UN)
        os.close(lock_fd)
    except IOError as x:
        eventlog.log_d(f'fail: {x}')


class UimodsLocker:
    def __init__(self, uimods: Union[Set[ph.UiMod], Set[ph.UiMod2]]):
        self._fd_list: List[int] = []
        try:
            self._fd_list = [self._lock(uimod) for uimod in uimods]
        except Exception:
            self.release()
            raise

    @staticmethod
    def _lock(uimod: Union[ph.UiMod, ph.UiMod2]) -> int:
        dir_path = os.path.dirname(uimod.lock_path)
        if not os.path.dirname(dir_path):
            os.makedirs(dir_path)

        lock_fd = os.open(uimod.lock_path, os.O_CREAT | os.O_TRUNC | os.O_WRONLY)
        try:
            # Request exclusive (EX) non-blocking (NB) advisory lock.
            fcntl.lockf(lock_fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return lock_fd

        except IOError as x:
            os.close(lock_fd)
            eventlog.log_d(f'fail: {x}')
            raise BusyError

    @staticmethod
    def _unlock(lock_fd: int) -> None:
        # Request exclusive (EX) non-blocking (NB) advisory lock.
        fcntl.lockf(lock_fd, fcntl.LOCK_UN)
        os.close(lock_fd)

    def release(self) -> None:
        # noinspection PyBroadException
        try:
            list(map(lambda lock_fd: self._unlock(lock_fd), self._fd_list))
        except Exception:
            eventlog.log_x()
            os.abort()


class CgiLock(object):
    def __init__(self, lock_file: str):
        self._lock_file = lock_file

    def __call__(self, fn, *extra_args, **extra_kwargs):
        return lambda *fn_args, **fn_kwargs: self._exec(fn, *fn_args, **fn_kwargs)

    # noinspection PyBroadException
    def _exec(self, fn, *fn_args, **fn_kwargs):
        if fs.try_lock_pid_file(self._lock_file):
            answer = fn(*fn_args, **fn_kwargs)
            if not fs.try_rm_pid_file(self._lock_file):
                eventlog.log_e(f'fail: rm pidfile <{self._lock_file}>')
            return answer

        else:
            eventlog.log_e(f'fail: BusyError (lock:<{self._lock_file}>)')
            raise BusyError
