from typing import List, Optional, Self, Set, Union

from edge.support import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap


# noinspection PyUnusedLocal
def ra_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_RADAR:
        return []

    try:
        ra_uimod, cam_uimod, fusion_uimod, location_uimod = (
            next(filter(lambda x: x.type == ph.UI_TYPE_RADAR, all_used_uimods)),
            next(filter(lambda x: x.type in [ph.UI_TYPE_MV_CAMERA], all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_LOCATION, all_used_uimods)))

        activate = any([
            uimod_to_newconf_map.get(location_uimod),
            uimod_to_newconf_map.get(cam_uimod)
        ])

        if forced or activate:
            return [RaCreator(entry_uimod, ra_uimod, cam_uimod, fusion_uimod, location_uimod)]

    except StopIteration:
        pass

    return []


class RaCreator(BaseCreator):
    def __init__(self, entry_uimod: ph.UiMod,
                 ra_uimod: ph.UiMod,
                 cam_uimod: ph.UiMod,
                 fusion_uimod: ph.UiMod,
                 location_uimod: ph.UiMod):

        BaseCreator.__init__(self)
        self._entry_uimod = entry_uimod
        self._ra_uimod = ra_uimod
        self._cam_uimod = cam_uimod
        self._fusion_uimod = fusion_uimod
        self._location_uimod = location_uimod
        self._edge_ra = dg.DgMod.create(dg.EDGE_TYPE_RADAR, ra_uimod.pid, ra_uimod.id)
        self._main_halt = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._edge_ra

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        from ._radar_uds import RaLocationEditor

        location_uiconf = uimod_to_newconf_map.get(self._location_uimod) or self._location_uimod.load_conf()
        location_pri = location_uiconf['ro']
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()
        spool_dir = ph.get_spool_dir(self._ra_uimod.pid)

        # send sms to module radar with new config
        if uimod_to_newconf_map.get(self._location_uimod) is not None:
            radar_lock_dirpath = f'/etc/edge/radar.{self._ra_uimod.pid}/lock'
            editor = RaLocationEditor.deserialize(location_uiconf, cam_uiconf, radar_lock_dirpath)
            editor.serialize(self._ra_uimod)

        driver_name = location_pri.get('driver', 'libdriver-smartmicro')

        cam_fps = cam_uiconf['rw']['videostream']['fps']
        offset_us = location_pri.get('offset', 0) * (1e6 / cam_fps)
        # create main.conf
        main_conf = {
            'api-server': {
                'endpoint': fs.join_path(spool_dir, 'ra{0}-ra{0}-uds'.format(self._ra_uimod.id))
            },
            'device': {
                'driver': f'/opt/edge/libexec/radar/{driver_name}.so',
                'endpoint': location_pri.get('uri', 'ummr0a:///dev/ttyS0'),
                'runtime': f'/etc/edge/radar.{self._ra_uimod.id}/runtime/'
            },
            'strobe-react': {
                'endpoint': fs.join_path(spool_dir, f'mg{self._cam_uimod.id}-ra{self._ra_uimod.id}-pipe'),
                'offset': int(offset_us)
            },
            'scene-event': {
                'endpoints': [
                    fs.join_path(spool_dir, f'ra{self._ra_uimod.id}-tr{self._fusion_uimod.id}-scene')
                ]
            },
            'monitoring': {
                'stat-file': fs.join_path(spool_dir, 'zabbix', f'ra{self._ra_uimod.id}'),
                'stat-tout': 30
            }
        }

        ra_zone = location_pri.get('ra-zone')
        if ra_zone is not None:
            main_conf['scene-event']['zone'] = ra_zone

        # create haltaction.conf
        halt_lines = location_pri.get('haltaction')
        if halt_lines is None:
            halt_lines = [
                '0:/opt/edge/venv/bin/python3.11 /opt/edge/bin/psuctl run led-radar color G blink 0HZ',

                '1:/opt/edge/venv/bin/python3.11 /opt/edge/bin/psuctl run led-radar color R blink 1HZ;'
                '/opt/edge/venv/bin/python3.11 /opt/edge/bin/psuctl run radar-reboot',

                '2:/opt/edge/venv/bin/python3.11 /opt/edge/bin/psuctl run led-radar color R blink 0HZ'
            ]

        halt_conf = '\n'.join(halt_lines) if halt_lines else ''

        self._main_halt = main_conf, halt_conf

    def complete(self) -> None:
        main_conf, halt_conf = self._main_halt

        edge_ra = self._edge_ra
        edge_ra.install()
        edge_ra.save_main_conf(main_conf)
        edge_ra.save_haltaction_conf(halt_conf)
        edge_ra.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        pass


class RaDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_RADAR:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RADAR, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
