from typing import List, Optional, Self, Set, Union

from edge.support import fs
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.creator import BaseCreator, UiModToConfMap


# noinspection PyUnusedLocal
def rg_get_creators(entry_uimod: Union[ph.UiMod, ph.UiMod2],
                    all_used_uimods: Set[Union[ph.UiMod, ph.UiMod2]],
                    uimod_to_newconf_map: UiModToConfMap,
                    forced: bool = False) -> List[BaseCreator]:

    if entry_uimod.type != ph.UI_TYPE_RTSP_CAMERA:
        return []

    try:
        cam_uimod, fusion_uimod, lpr_uimod, rl_uimods, railroad_crossing_uimods = (
            next(filter(lambda x: x.type == ph.UI_TYPE_RTSP_CAMERA, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_FUSION, all_used_uimods)),
            next(filter(lambda x: x.type == ph.UI_TYPE_LPR, all_used_uimods)),
            [x for x in all_used_uimods if x.type == ph.UI_TYPE_RED_LIGHT],
            [x for x in all_used_uimods if x.type == ph.UI_TYPE_RAILROAD_CROSSING])

        activate = any([
            uimod_to_newconf_map.get(cam_uimod),
            any([uimod_to_newconf_map.get(x) for x in rl_uimods]),
            any([uimod_to_newconf_map.get(x) for x in railroad_crossing_uimods])
        ])

        return [RgCreator(cam_uimod, lpr_uimod, fusion_uimod, rl_uimods, railroad_crossing_uimods)]

    except StopIteration:
        pass

    return []


class RgCreator(BaseCreator):
    def __init__(self, cam_uimod: ph.UiMod,
                 lpr_uimod: ph.UiMod,
                 fusion_uimod: ph.UiMod,
                 rl_uimods: List[ph.UiMod],
                 railroad_crossing_uimods: List[ph.UiMod]):

        BaseCreator.__init__(self)
        self._cam_uimod = cam_uimod
        self._lpr_uimod = lpr_uimod
        self._fusion_uimod = fusion_uimod
        self._rl_uimods = rl_uimods
        self._railroad_crossing_uimods = railroad_crossing_uimods
        self._rg_dgmod = dg.DgMod.create(
            dg.EDGE_TYPE_RTSP_GRABBER, cam_uimod.pid, cam_uimod.id)
        self._main_halt = None

    @property
    def dgmod(self) -> dg.DgMod:
        return self._rg_dgmod

    def prepare(self, uimod_to_newconf_map: UiModToConfMap) -> None:
        ui_pid = self._cam_uimod.pid
        cam_uiconf = uimod_to_newconf_map.get(self._cam_uimod) or self._cam_uimod.load_conf()

        use_udp = False
        transport = cam_uiconf['rw']['conf']['video'].get('transport', ['RTP/AVP/UDP', 'RTP/AVP/TCP'])
        if 'RTP/AVP/UDP' in transport:
            use_udp = True

        skip_rate = cam_uiconf['rw']['conf']['video'].get('skip-rate', 1) - 1
        spool_dir = ph.get_spool_dir(ui_pid)

        ui_format_to_edge_format = {
            'video/H264': 'h264',
            'video/JPEG': 'jpeg'
        }
        use_gpu = 'hw/auto' == cam_uiconf['rw']['conf']['video'].get('decoder', 'hw/auto')
        # create main.conf

        mono8_consumers = []
        raw_consumers = []
        main_conf = {
            'stats': {
                'filepath': fs.join_path(spool_dir, 'zabbix', 'vs0'),
                'interval': 5
            },
            'dec-source': {
                'conf': {
                    'target-w': 0,
                    'target-h': 0,
                    'skiprate': skip_rate,
                    'prefer-gpu': use_gpu,
                    'with-deep-notify': False,
                    'wdt': 0
                }
            },
            'enc-streamer': [
                # {
                #     'type': 'bus-streamer',
                #     'conf': {
                #         'format': 'raw',
                #         'blob-writer': {
                #             'file-dir': '/var/spool/edge/rg0-rg0-store/',
                #             'file-count': [50, 60]
                #         },
                #         'args-writer': [
                #             '/var/spool/edge/rg0-lr0-scene/'
                #         ]
                #     }
                # },
            ],
            'dec-streamer': [
            ]
        }

        # TODO: if http-source
        flexwatch_conf = cam_uiconf['ro'].get('flexwatch-conf')
        if flexwatch_conf:
            main_conf['enc-source'] = {
                'type': 'flexwatch',
                'conf': flexwatch_conf
            }
        else:
            main_conf['enc-source'] = {
                'type': 'rtsp',
                'conf': {
                    'url': cam_uiconf['rw']['conf']['url'],
                    'video-format': ui_format_to_edge_format.get(cam_uiconf['rw']['conf']['video']['format']),
                    'allow-udp-unicast': use_udp,
                    'allow-udp-multicast': use_udp,
                    'with-deep-notify': False,
                    'wdt': 0
                }
            }

        rgbs = cam_uiconf['ro'].get('rgbs', False)
        mcgst_scene = cam_uiconf['ro'].get('mcgst-scene', True)

        if rgbs:
            main_conf['dec-streamer'].append({
                'type': 'bsv-streamer',
                'conf': {
                    'format': 'mjpeg',
                    'host': int(self._rg_dgmod.id),
                    'chan': '0'
                }
            })

        lr_endpoint = fs.join_path(spool_dir, f'rg{self._cam_uimod.id}-lr{self._lpr_uimod.id}-scene')
        tr_endpoint = fs.join_path(spool_dir, f'rg{self._cam_uimod.id}-tr{self._fusion_uimod.id}-scene')
        mc_endpoint = fs.join_path(spool_dir, f'mg{self._cam_uimod.id}-mcgst{self._cam_uimod.id}-scene')

        if cam_uiconf['rw']['conf']['video'].get('grayscale', False):
            consumers = mono8_consumers
        else:
            consumers = raw_consumers

        consumers.append(lr_endpoint)
        consumers.append(tr_endpoint)
        if mcgst_scene and not rgbs:
            consumers.append(mc_endpoint)

        # add mg to rl path
        tl_uimod = None
        if len(self._rl_uimods) == 1:
            tl_uimod = self._rl_uimods[0]
        elif len(self._railroad_crossing_uimods) == 1:
            tl_uimod = self._railroad_crossing_uimods[0]

        if tl_uimod is not None:
            red_light_jnode = uimod_to_newconf_map.get(tl_uimod) or tl_uimod.load_conf()
            vendor = red_light_jnode['ro'].get('vendor', 'zonechecker')
            if vendor == 'zonechecker':
                endpoint = fs.join_path(spool_dir, f'mg{self._cam_uimod.id}-zc{tl_uimod.id}-scene')
                consumers.append(endpoint)

        if mono8_consumers:
            main_conf['dec-streamer'].append({
                'type': 'bus-streamer',
                'conf': {
                    'format': 'mono8',
                    'blob-writer': {
                        'file-dir': fs.join_path(spool_dir, 'rg{0}-rg{0}-store-mono8'.format(self._cam_uimod.id)),
                        'file-count': [40, 45]
                    },
                    'args-writer': mono8_consumers
                }
            })

        if raw_consumers:
            main_conf['dec-streamer'].append({
                'type': 'bus-streamer',
                'conf': {
                    'format': 'raw',
                    'blob-writer': {
                        'file-dir': fs.join_path(spool_dir, 'rg{0}-rg{0}-store-raw'.format(self._cam_uimod.id)),
                        'file-count': [40, 45]
                    },
                    'args-writer': raw_consumers
                }
            })

        # create haltaction.conf
        def_halt_list = [
            '0:/opt/edge/venv/bin/python3.11 /opt/edge/bin/psuctl run led-camera color G blink 0HZ',
            '1:/opt/edge/venv/bin/python3.11 /opt/edge/bin/psuctl run led-camera color R blink 1HZ',
            '2:/opt/edge/venv/bin/python3.11 /opt/edge/bin/psuctl run led-camera color R blink 0HZ'
        ]
        halt_list = cam_uiconf['ro'].get('haltaction', def_halt_list)
        if halt_list:
            halt_conf = '\n'.join(halt_list)

        else:
            halt_conf = ''

        self._main_halt = main_conf, halt_conf

    def complete(self) -> None:
        main_conf, halt_conf = self._main_halt  # type: dict, str

        edge_rg = self._rg_dgmod
        edge_rg.install()
        edge_rg.save_main_conf(main_conf)
        edge_rg.save_haltaction_conf(halt_conf)
        edge_rg.save_setupstore_conf(active=True)

    def rollback(self) -> None:
        pass

    def notify(self) -> None:
        dg.LifeCycle.restart_dgmod(self.dgmod.type, self.dgmod.id)


class RgDeleter:
    def __init__(self, dgmod: dg.DgMod):
        self._dgmod = dgmod

    @classmethod
    def create(cls, entry_uimod: ph.UiMod) -> Optional[Self]:
        if entry_uimod.type == ph.UI_TYPE_RTSP_CAMERA:
            dgmod = dg.DgMod.create(dg.EDGE_TYPE_RTSP_GRABBER, entry_uimod.pid, entry_uimod.id)
            return cls(dgmod)

    def run(self) -> None:
        self._dgmod.uninstall()
