import ctypes
import functools
import io
import os
import socket
import tarfile
import ujson

from typing import Callable, Dict, List, Optional

from edge import eventlog
from edge.services.types.about import IAboutService
from edge.services.types.tz import ITzService
from edge.support.bs import BsError
from edge.support.fs import silent_mkdir_r
from edge.support.queue import JsonFileQueue
from edge.types.blob import MemBlob, Blob, StructBlobMeta
from edge.types.exceptions import BufferSizeError

from speedcontrol.events.event import Event
from speedcontrol.events.jnode.event import encode__event, encode__event_ua
from speedcontrol.services.types.bs import IBsService
from speedcontrol.services.types.event import EventServiceConf, IEventService
from speedcontrol.services.types.sign import SignError, ISignService
from speedcontrol.services.types.stats import IEventCounters
from speedcontrol.services.types.systest import ISystestService


def _make_fact(event_jnode: Dict,
               args_blob_ref: str,
               args_ua_blob_ref: Optional[str] = None,
               args_ua_blob_sign: Optional[bool] = False) -> Dict:

    seatbelt_event = event_jnode['type'] == 'seatbelt'

    def add_arch_file(name: str, altref: Optional[str] = None) -> None:
        arch_file = {
            'ref': name,
            'name': name
        }

        if altref:
            arch_file['altref'] = altref

        arch_files.append(arch_file)

    def analyze_measurements() -> None:
        measurements = event_jnode['measurements']

        for measurement in measurements:
            waypoints = measurement['waypoints']
            best_images = waypoints['best'].get('images', {})
            plate_images = best_images.get('license-plates', [])

            # thumb.jpeg
            thumb_im_name = best_images.get('thumb', {}).get('name')
            if thumb_im_name is not None:
                add_arch_file(thumb_im_name, 'thumb.jpeg')

            # scene.jpeg
            scene_im_name = best_images.get('scene', {}).get('name')
            if scene_im_name is not None:
                add_arch_file(scene_im_name, 'scene.jpeg')

            # vehicle.jpeg
            vehicle_im_name = best_images.get('vehicle', {}).get('name')
            if vehicle_im_name is not None:
                add_arch_file(vehicle_im_name, 'vehicle.jpeg')

            # overview.jpeg
            overview_im_name = best_images.get('overview', {}).get('name', None)
            if overview_im_name is not None:
                add_arch_file(overview_im_name, 'overview.jpeg')

            # plate.jpeg
            list(map(lambda plate_im: add_arch_file(plate_im['name'], 'plate.jpeg'), plate_images))

    def analyze_evidence() -> None:
        e_jnode = event_jnode.get('evidence', {})
        em_jnode = e_jnode.get('measurements')
        if not em_jnode:
            return

        emw_jnodes = em_jnode.get('waypoints', {})
        for waypoint in emw_jnodes.values():
            if type(waypoint) != list:
                waypoint = [waypoint]

            for w_jnode in waypoint:
                wi_jnodes = w_jnode.get('images', {})
                if seatbelt_event:
                    plate_image = wi_jnodes.get('plate', {}).get('name')
                    add_arch_file(plate_image, 'plate.jpeg')

                    # thumb.jpeg
                    thumb_im_name = wi_jnodes.get('thumb', {}).get('name')
                    if thumb_im_name is not None:
                        add_arch_file(thumb_im_name, 'thumb.jpeg')

                    # scene.jpeg
                    scene_im_name = wi_jnodes.get('scene', {}).get('name')
                    if scene_im_name is not None:
                        add_arch_file(scene_im_name, 'scene.jpeg')

                    # vehicle.jpeg
                    vehicle_im_name = wi_jnodes.get('vehicle', {}).get('name')
                    if vehicle_im_name is not None:
                        add_arch_file(vehicle_im_name, 'vehicle.jpeg')

                else:
                    for wi_jnode in wi_jnodes.values():
                        add_arch_file(wi_jnode['name'])

        emv_jnodes = em_jnode.get('videos', [])
        for emv_jnode in emv_jnodes:
            add_arch_file(emv_jnode['name'])

    arch_files = []
    add_arch_file(args_blob_ref, 'args.json')

    if args_ua_blob_ref is not None:
        add_arch_file(args_ua_blob_ref, 'args-ua.json')

        if args_ua_blob_sign:
            sign_name = os.path.splitext(args_ua_blob_ref)[0] + '.sign'
            cert_name = os.path.splitext(args_ua_blob_ref)[0] + '.crt'
            add_arch_file(sign_name, 'args-ua.sign')
            add_arch_file(cert_name, 'args-ua.crt')

    analyze_measurements()
    analyze_evidence()

    jnode = {
        'args': event_jnode,
        'dist': arch_files
    }

    return jnode


class _EventService(IEventService):
    def __init__(self, twisted_reactor,
                 counters: IEventCounters,
                 tz_service: ITzService,
                 bs_service: IBsService,
                 systest_service: ISystestService,
                 about_service: IAboutService,
                 sign_service: ISignService,
                 json_q_list: List[JsonFileQueue],
                 args: EventServiceConf):

        IEventService.__init__(self)
        self._reactor = twisted_reactor
        self._counters = counters

        self._tz_service = tz_service
        self._bs_service = bs_service
        self._systest_service = systest_service
        self._about_service = about_service
        self._sign_service = sign_service
        self._json_q_list = json_q_list
        self._devices_jnode = self._get_devices(about_service)

        self._args_ua_required = args.ua_json_required
        self._local_arch_enabled = args.local_arch_enabled
        self._local_arch_base_dir = args.local_arch_base_dir

        self._hash_buffer_size = 100

    @staticmethod
    def _get_devices(about_service: IAboutService) -> List[Dict]:
        device = {'hostname': socket.gethostname()}
        hw = about_service.hw_info
        if any(hw):
            device['hw'] = hw

        sw = about_service.sw_info
        if sw:
            device['sw'] = sw

        sw_name = about_service.sw_name
        if sw_name:
            device['sw-name'] = sw_name

        return [device]

    def _hash_blob(self, blob: Blob) -> None:
        hash_buffer = ctypes.create_string_buffer(b'', size=self._hash_buffer_size)
        hash_size = 0
        try:
            hash_size = self._sign_service.hash_blob(blob.file_data, hash_buffer)
        except BufferSizeError as x:
            self._hash_buffer_size = x.min_size
            hash_buffer = ctypes.create_string_buffer(b'', size=self._hash_buffer_size)
            hash_size = self._sign_service.hash_blob(blob.file_data, hash_buffer)

        except SignError:
            eventlog.log_e('fail: hash <profile={}>'.format(blob.profile))

        if hash_size > 0:
            blob.hash = self._sign_service.hash_type, hash_buffer.raw[:hash_size]

    def _emit2(self, event_args, blobs_to_test, args_blob, args_ua_blob, media_blobs, args_jnode, args_ua_signed):
        def get_tar_path(base_dir: str, ts: int) -> str:
            d = self._tz_service.map_ts_usec_to_datetime_local(ts)
            path = os.path.join(base_dir, str(d.year), str(d.month).zfill(2),
                                str(d.day).zfill(2), str(d.hour).zfill(2),
                                str(d.minute).zfill(2))
            return path

        def export_event_to_gzip() -> None:
            ts = event_args.ts
            filedir = get_tar_path(self._local_arch_base_dir, ts)
            silent_mkdir_r(filedir)
            args_blob_id = args_blob.name.split('_')[1].split(".")[0]
            tar_name = os.path.join(filedir, '{}.tar'.format(args_blob_id))

            with tarfile.open(name=tar_name, mode='w') as tar:
                args_info = tarfile.TarInfo('args.json')
                args_info.size = len(args_blob.file_data)
                tar.addfile(args_info, io.BytesIO(args_blob.file_data))

                if args_ua_blob is not None:
                    args_info = tarfile.TarInfo('args-ua.json')
                    args_info.size = len(args_ua_blob.file_data)
                    tar.addfile(args_info, io.BytesIO(args_ua_blob.file_data))

                for media_blob in media_blobs:
                    if media_blob.file_path is None:
                        tar_info = tarfile.TarInfo(media_blob.name)
                        tar_info.size = len(media_blob.file_data)
                        tar.addfile(tar_info, io.BytesIO(media_blob.file_data))
                    else:
                        tar.add(media_blob.file_path, media_blob.name)

        try:

            if args_ua_blob:
                self._bs_service.save(event_args.sender, args_ua_blob)
                self._hash_blob(args_ua_blob)
                blobs_to_test.append(args_ua_blob)

                fact_jnode = _make_fact(args_jnode, args_blob.name, args_ua_blob.name, args_ua_signed)
            else:
                fact_jnode = _make_fact(args_jnode, args_blob.name)

            skip_event = False
            for blob in blobs_to_test:
                blob_id = blob.name.split('_')[1].split('.')[0]
                blob_size = self._bs_service.get_size(event_args.sender, blob_id, blob.profile)
                blob_size2 = os.stat(blob.file_path).st_size if blob.file_path else len(blob.file_data)
                if blob_size != blob_size2:
                    skip_event = True
                    eventlog.log_i('blob_id {} / {} profile {:12s} blob_size {:7d} / {:7d} ALARM'.format(
                        blob_id, hex(int(blob_id)), blob.profile, blob_size, blob_size2))

            if not skip_event:
                if self._local_arch_enabled:
                    export_event_to_gzip()

                list(map(lambda json_q: json_q.add_jnode(fact_jnode, args_blob.name), self._json_q_list))
                self._counters.inc_fact()
            else:
                eventlog.log_w('fail: skip event due to ALARM-s')

        except BsError:
            eventlog.log_w('fail: save-blob (skip event)')

        except Exception:
            eventlog.log_x()

    def emit(self, event: Event) -> bool:
        def handle_sign(
                event_args2,
                blobs_to_test2,
                args_blob2,
                args_ua_blob2,
                media_blobs2,
                args_jnode2,
                blob_data: Optional[str],
                sign_data: Optional[str]):

            args_ua_signed = sign_data is not None

            if args_ua_signed:
                if blob_data != args_ua_blob2.file_data:
                    eventlog.log_w("fail: blob_data != args_ua_blob.file_data")

                crt = self._sign_service.cert
                if crt:
                    args_ua_blob2.meta = StructBlobMeta(sign_data, len(sign_data), crt)
                else:
                    args_ua_blob2.meta = StructBlobMeta(sign_data, len(sign_data))

            self._emit2(event_args2, blobs_to_test2, args_blob2, args_ua_blob2, media_blobs2,
                        args_jnode2, args_ua_signed)

        def sign_blob(blob_to_sign: Blob, callback: Callable[[Optional[str], Optional[str]], None]) -> Optional[int]:
            try:
                return self._sign_service.sign_blob(blob_to_sign.file_data, callback)

            except SignError:
                pass

            except Exception:
                eventlog.log_x()

            callback(None, None)

        # noinspection PyBroadException
        try:
            # save media
            media_blobs = event.get_blobs()
            list(map(lambda m_blob: self._bs_service.save(event.sender, m_blob), media_blobs))
            list(map(self._hash_blob, media_blobs))

            # save args.json
            args_jnode = encode__event(event, self._tz_service)
            self._systest_service.serialize(args_jnode)
            args_jnode['origin']['devices'] = self._devices_jnode
            args_blob = MemBlob('args', event.ts, bytes(ujson.dumps(args_jnode), 'ascii'), file_ext='json')

            self._bs_service.save(event.sender, args_blob)
            self._hash_blob(args_blob)

            # save args-ua.json
            blobs_to_test = media_blobs + [args_blob]
            args_ua_blob = None
            if self._args_ua_required:
                args_ua_jnode = encode__event_ua(event, self._tz_service, self._about_service)
                args_ua_blob = MemBlob(
                    'args', event.ts, bytes(ujson.dumps(args_ua_jnode), 'ascii'), file_ext='json')

                if event.type in [Event.TYPE_BUS_LANE,
                                  Event.TYPE_SPOT_SPEED,
                                  Event.TYPE_SECTION_SPEED,
                                  Event.TYPE_WRONG_WAY,
                                  Event.TYPE_RED_LIGHT,
                                  Event.TYPE_PARKING,
                                  Event.TYPE_WANTED]:

                    handler = functools.partial(
                        handle_sign, event, blobs_to_test,
                        args_blob, args_ua_blob, media_blobs, args_jnode)

                    sign_blob(args_ua_blob, handler)
                else:
                    self._emit2(event, blobs_to_test, args_blob, args_ua_blob, media_blobs, args_jnode, False)

            else:
                self._emit2(event, blobs_to_test, args_blob, args_ua_blob, media_blobs, args_jnode, False)

        except BsError:
            eventlog.log_w('fail: save-blob (skip event)')
            return False

        except Exception:
            eventlog.log_x()
            return False

        return True


def init_service(
        twisted_reactor,
        counters: IEventCounters,
        args: EventServiceConf,
        tz_service: ITzService,
        bs_service: IBsService,
        systest_service: ISystestService,
        about_service: IAboutService,
        sign_service: ISignService) -> IEventService:

    json_q_list = [JsonFileQueue(json_q_args) for json_q_args in args.json_queue_list]
    return _EventService(twisted_reactor, counters, tz_service, bs_service, systest_service,
                         about_service, sign_service, json_q_list, args)
